==> pdb-tools-2.0.1.tar.gz: pdb-tools-2.0.1/pdbtools/pdb_delinsertion.py <==
#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2018 João Pedro Rodrigues
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Deletes insertion codes in a PDB file, shifting the residue numbering of
downstream residues. Allows for picking specific residues to delete insertion
codes for.

Usage:
    python pdb_delinsertion.py [-<option>] <pdb file>

Example:
    python pdb_delinsertion.py 1CTF.pdb  # delete ALL insertion codes EVERYWHERE
    python pdb_delinsertion.py -A99,B12 1CTF.pdb  # deletes ins. codes for residue
                                              # 99 of chain A and 12 of chain B.

This program is part of the `pdb-tools` suite of utilities and should not be
distributed isolatedly. The `pdb-tools` were created to quickly manipulate PDB
files using the terminal, and can be used sequentially, with one tool streaming
data to another. They are based on old FORTRAN77 code that was taking too much
effort to maintain and compile. RIP.
"""

import os
import sys

__author__ = "Joao Rodrigues"
__email__ = "j.p.g.l.m.rodrigues@gmail.com"


def check_input(args):
    """Checks whether to read from stdin/file and validates user input/options.
    """

    # Defaults
    option = ''
    fh = sys.stdin  # file handle

    if not len(args):
        # Reading from pipe with default option
        if sys.stdin.isatty():
            sys.stderr.write(__doc__)
            sys.exit(1)

    elif len(args) == 1:
        # One of two options: option & Pipe OR file & default option
        if args[0].startswith('-'):
            option = args[0][1:]
            if sys.stdin.isatty():  # ensure the PDB data is streamed in
                emsg = 'ERROR!! No data to process!\n'
                sys.stderr.write(emsg)
                sys.stderr.write(__doc__)
                sys.exit(1)

        else:
            if not os.path.isfile(args[0]):
                emsg = 'ERROR!! File not found or not readable: \'{}\'\n'
                sys.stderr.write(emsg.format(args[0]))
                sys.stderr.write(__doc__)
                sys.exit(1)

            fh = open(args[0], 'r')

    elif len(args) == 2:
        # Two options: option & File
        if not args[0].startswith('-'):
            emsg = 'ERROR! First argument is not an option: \'{}\'\n'
            sys.stderr.write(emsg.format(args[0]))
            sys.stderr.write(__doc__)
            sys.exit(1)

        if not os.path.isfile(args[1]):
            emsg = 'ERROR!! File not found or not readable: \'{}\'\n'
            sys.stderr.write(emsg.format(args[1]))
            sys.stderr.write(__doc__)
            sys.exit(1)

        option = args[0][1:]
        fh = open(args[1], 'r')

    else:  # Whatever ...
        sys.stderr.write(__doc__)
        sys.exit(1)

    # Validate option
    option_list = [o for o in option.split(',') if o.strip()]
    if option_list:
        for o in option_list:
            if len(o) < 2 or not o[1:].isdigit():
                emsg = 'ERROR!! Option invalid: \'{}\''
                sys.stderr.write(emsg.format(o))
                sys.exit(1)

    return (option_list, fh)


def delete_insertions(fhandle, option_list):
    """Deletes insertion codes (at specific residues).

    By default, removes ALL insertion codes on ALL residues. Also bumps the
    residue numbering of residues downstream of each insertion.
    """

    if not option_list:  # Delete all
        has_option = False
    else:  # turn into set
        option_set = set(option_list)
        has_option = True

    # Keep track of residue numbering
    # Keep track of residues read (chain, resname, resid)
    offset = 0
    prev_resi = None
    seen_ids = set()
    clean_icode = False
    records = ('ATOM', 'HETATM', 'ANISOU', 'TER')
    for line in fhandle:

        if line.startswith(records):
            res_uid = line[17:26]  # resname, chain, resid
            id_res = line[21] + line[22:26].strip()  # A99, B12
            icode = line[26]

            # unfortunately, this is messy but not all PDB files follow a nice
            # order of ' ', 'A', 'B', ... when it comes to insertion codes..
            if prev_resi != res_uid:  # new residue
                # Have we seen this chain + resid combination
                # catch insertions WITHOUT icode ('A' ... ' ' ... 'B')
                if id_res in seen_ids:
                    # Should we do something about it?
                    if has_option and id_res not in option_set:
                        clean_icode = False
                    else:
                        clean_icode = True
                        line = line[:26] + ' ' + line[27:]  # clear icode
                        offset += 1
                # Do we have an explicit icode?
                elif icode != ' ':
                    if has_option and id_res not in option_set:
                        pass
                    else:
                        if id_res in seen_ids:  # never saw this, do not offset!
                            offset += 1
                        clean_icode = True
                        line = line[:26] + ' ' + line[27:]  # clear icode
                else:
                    clean_icode = False

                prev_resi = res_uid

            if clean_icode:
                line = line[:26] + ' ' + line[27:]

            resid = int(line[22:26]) + offset
            line = line[:22] + str(resid).rjust(4) + line[26:]
            seen_ids.add(id_res)

            # Reset offset on TER
            if line.startswith('TER'):
                offset = 0

        yield line


def main():
    # Check Input
    option_list, pdbfh = check_input(sys.argv[1:])

    # Do the job
    new_pdb = delete_insertions(pdbfh, option_list)

    try:
        _buffer = []
        _buffer_size = 5000  # write N lines at a time
        for lineno, line in enumerate(new_pdb):
            if not (lineno % _buffer_size):
                sys.stdout.write(''.join(_buffer))
                _buffer = []
            _buffer.append(line)

        sys.stdout.write(''.join(_buffer))
        sys.stdout.flush()
    except IOError:
        # This is here to catch Broken Pipes
        # for example to use 'head' or 'tail' without
        # the error message showing up
        pass

    # last line of the script
    # We can close it even if it is sys.stdin
    pdbfh.close()
    sys.exit(0)


if __name__ == '__main__':
    main()

==> pdb-tools-2.0.3.tar.gz: pdb-tools-2.0.3/pdbtools/pdb_delinsertion.py <==
#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2018 João Pedro Rodrigues
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Deletes insertion codes in a PDB file.

Deleting an insertion code shifts the residue numbering of downstream
residues. Allows for picking specific residues to delete insertion codes for.

Usage:
    python pdb_delinsertion.py [-<option>] <pdb file>

Example:
    python pdb_delinsertion.py 1CTF.pdb  # delete ALL insertion codes
    python pdb_delinsertion.py -A9,B12 1CTF.pdb  # deletes ins. codes for res
                                                 # 9 of chain A and 12 of chain B.

This program is part of the `pdb-tools` suite of utilities and should not be
distributed isolatedly. The `pdb-tools` were created to quickly manipulate PDB
files using the terminal, and can be used sequentially, with one tool streaming
data to another. They are based on old FORTRAN77 code that was taking too much
effort to maintain and compile. RIP.
"""

import os
import sys

__author__ = "Joao Rodrigues"
__email__ = "j.p.g.l.m.rodrigues@gmail.com"


def check_input(args):
    """Checks whether to read from stdin/file and validates user input/options.
    """

    # Defaults
    option = ''
    fh = sys.stdin  # file handle

    if not len(args):
        # Reading from pipe with default option
        if sys.stdin.isatty():
            sys.stderr.write(__doc__)
            sys.exit(1)

    elif len(args) == 1:
        # One of two options: option & Pipe OR file & default option
        if args[0].startswith('-'):
            option = args[0][1:]
            if sys.stdin.isatty():  # ensure the PDB data is streamed in
                emsg = 'ERROR!! No data to process!\n'
                sys.stderr.write(emsg)
                sys.stderr.write(__doc__)
                sys.exit(1)

        else:
            if not os.path.isfile(args[0]):
                emsg = 'ERROR!! File not found or not readable: \'{}\'\n'
                sys.stderr.write(emsg.format(args[0]))
                sys.stderr.write(__doc__)
                sys.exit(1)

            fh = open(args[0], 'r')

    elif len(args) == 2:
        # Two options: option & File
        if not args[0].startswith('-'):
            emsg = 'ERROR! First argument is not an option: \'{}\'\n'
            sys.stderr.write(emsg.format(args[0]))
            sys.stderr.write(__doc__)
            sys.exit(1)

        if not os.path.isfile(args[1]):
            emsg = 'ERROR!! File not found or not readable: \'{}\'\n'
            sys.stderr.write(emsg.format(args[1]))
            sys.stderr.write(__doc__)
            sys.exit(1)

        option = args[0][1:]
        fh = open(args[1], 'r')

    else:  # Whatever ...
        sys.stderr.write(__doc__)
        sys.exit(1)

    # Validate option
    option_list = [o for o in option.split(',') if o.strip()]
    if option_list:
        for o in option_list:
            if len(o) < 2 or not o[1:].isdigit():
                emsg = 'ERROR!! Option invalid: \'{}\''
                sys.stderr.write(emsg.format(o))
                sys.exit(1)

    return (option_list, fh)


def delete_insertions(fhandle, option_list):
    """Deletes insertion codes (at specific residues).

    By default, removes ALL insertion codes on ALL residues. Also bumps the
    residue numbering of residues downstream of each insertion.
    """

    if not option_list:  # Delete all
        has_option = False
    else:  # turn into set
        option_set = set(option_list)
        has_option = True

    # Keep track of residue numbering
    # Keep track of residues read (chain, resname, resid)
    offset = 0
    prev_resi = None
    seen_ids = set()
    clean_icode = False
    records = ('ATOM', 'HETATM', 'ANISOU', 'TER')
    for line in fhandle:

        if line.startswith(records):
            res_uid = line[17:26]  # resname, chain, resid
            id_res = line[21] + line[22:26].strip()  # A99, B12
            icode = line[26]

            # unfortunately, this is messy but not all PDB files follow a nice
            # order of ' ', 'A', 'B', ... when it comes to insertion codes..
            if prev_resi != res_uid:  # new residue
                # Have we seen this chain + resid combination
                # catch insertions WITHOUT icode ('A' ... ' ' ... 'B')
                if id_res in seen_ids:
                    # Should we do something about it?
                    if has_option and id_res not in option_set:
                        clean_icode = False
                    else:
                        clean_icode = True
                        line = line[:26] + ' ' + line[27:]  # clear icode
                        offset += 1
                # Do we have an explicit icode?
                elif icode != ' ':
                    if has_option and id_res not in option_set:
                        pass
                    else:
                        if id_res in seen_ids:  # never saw this, do not offset!
                            offset += 1
                        clean_icode = True
                        line = line[:26] + ' ' + line[27:]  # clear icode
                else:
                    clean_icode = False

                prev_resi = res_uid

            if clean_icode:
                line = line[:26] + ' ' + line[27:]

            resid = int(line[22:26]) + offset
            line = line[:22] + str(resid).rjust(4) + line[26:]
            seen_ids.add(id_res)

            # Reset offset on TER
            if line.startswith('TER'):
                offset = 0

        yield line


def main():
    # Check Input
    option_list, pdbfh = check_input(sys.argv[1:])

    # Do the job
    new_pdb = delete_insertions(pdbfh, option_list)

    try:
        _buffer = []
        _buffer_size = 5000  # write N lines at a time
        for lineno, line in enumerate(new_pdb):
            if not (lineno % _buffer_size):
                sys.stdout.write(''.join(_buffer))
                _buffer = []
            _buffer.append(line)

        sys.stdout.write(''.join(_buffer))
        sys.stdout.flush()
    except IOError:
        # This is here to catch Broken Pipes
        # for example to use 'head' or 'tail' without
        # the error message showing up
        pass

    # last line of the script
    # We can close it even if it is sys.stdin
    pdbfh.close()
    sys.exit(0)


if __name__ == '__main__':
    main()
